"""
This module contains ShapingObject implementation.

There is no need to import this module directly. Use the methods of
the L{Ruleset object <ruleset.Ruleset>} to create and manipulate
ShapingObjects.

    >>> so = r.shapingobject_add('Test Shaping',
                                 inbound=[(0, 0, 1024)], # 1024 kbyte/s inbound
                                 outbound=[(0, 0, 512)]) # 512 kbyte/s outbound

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

import plutil

from packetlogic2.exceptions import PLDBUnsupportedInFirmware

NONEXISTANT_SPLIT_TYPE = object()

def _shapingobjectlist(pldb):
	cid = pldb.session.shaping_object_list()
	return plutil._do(pldb, cid, plutil._ListCB(pldb, pldb.session.shaping_object_list_handle, ShapingObject))

def _shapingobjectadd(_pldb, name, vbs_hours, inbound, outbound, bidir, split, max_connections, subscriber_netobject, flags):
	inbound = _makenewformat(inbound)
	outbound = _makenewformat(outbound)
	bidir = _makenewformat(bidir)
	kwargs = locals().copy()

	flags = ShapingObjectFlags(flags)

	cid = _pldb.session.shaping_object_add(name, (inbound, outbound, bidir), split, max_connections, subscriber_netobject, int(flags))
	(id, tstamp) = plutil._do(_pldb, cid, plutil._AddCB(_pldb, _pldb.session.shaping_object_add_handle))
	res = ShapingObject(id = id,
			    creation_date = tstamp,
			    is_global = False,
			    modification_date = tstamp,
			    creator = _pldb.user,
			    modified_by = _pldb.user,
			    systemid = _pldb.systemid,
			    **kwargs)
	return res

def _shapingobjectupdate(_pldb, obj, name, vbs_hours, inbound, outbound, bidir, split, max_connections, subscriber_netobject, intflags):
	#v = locals()
	cid = _pldb.session.shaping_object_modify(obj.id, name, (inbound, outbound, bidir), split, max_connections, subscriber_netobject, intflags)
	tstamp = plutil._do(_pldb, cid, plutil._AddCB(_pldb, _pldb.session.shaping_object_modify_handle))
	obj._prop.modification_date = tstamp
	obj._prop.modified_by = _pldb.user
	#[setattr(obj._prop, var, val) for var, val in v.iteritems() if var not in ["_pldb", "obj", "intflags"]]

def _rename(pldb, obj, newname):
	cid = pldb.session.shaping_object_rename(obj.id, newname)
	return plutil._do(pldb, cid, plutil._AddCB(pldb, pldb.session.shaping_object_rename_handle))

def _remove(pldb, obj):
	cid = pldb.session.shaping_object_remove(obj.id)
	return plutil._do(pldb, cid, plutil._AddCB(pldb, pldb.session.shaping_object_remove_handle))

_Prop_ShapingObjectFlags = plutil.Prop(brown=bool, host_fairness=bool, counter=bool)
class ShapingObjectFlags(plutil.BaseFlags):
	""" Used internaly to represent the flags on a L{ShapingObject}.

	Use L{ShapingObjects.set_flags} to change the flags.
	"""
	__slots__ = ('_prop')
	_Prop_class = _Prop_ShapingObjectFlags

	def _set_flags_other_type(self, intflags):
		""" Support bitmap int to set flags (when creted from API list)."""
		if isinstance(intflags, int):
			self._prop.update(brown=intflags & 0x01, host_fairness=intflags & 0x02, counter=intflags & 0x10)
		else:
			plutil.BaseFlags._set_flags_other_type(self, intflags)

	def __int__(self):
		return self._prop.brown * 0x01 + \
		    self._prop.host_fairness * 0x02 + \
		    self._prop.counter * 0x10

	brown = property(lambda self: self._prop.brown, doc="Enable the BROWN Shaping Algorithm")
	host_fairness = property(lambda self: self._prop.host_fairness, doc="Enable fairness between hosts")
	counter = property(lambda self: self._prop.counter, doc="Enable byte counter")
	blue = property(lambda self: False, doc="Enable BLUE AQM (requires firmware 14.0 or newer)")
	virtual_queueing = property(lambda self: False, doc="Enable virtual queueing (requires firmware 14.0 or newer)")
	codel = property(lambda self: False, doc="Enable CoDel AQM (requires firmware 14.0 or newer)")
	wfq = property(lambda self: False, doc="Enable WFQ configured on the wfq property on the shapingobject (requires firmware 14.0 or newer)")



def _makeoldformat(limits, name):
	"""Convert limits to old format"""
	import warnings
	warnings.warn("Use of property '%s' is deprecated, use 'limits.%s' instead" % (name, name),
	              DeprecationWarning, stacklevel=2)
	return [x[:3] for x in limits]

def _makenewformat(limits):
	r = []
	for l in limits:
		if len(l) == 3:
			l = l + (0,0,0)
		r.append(l)
	return r

class ShapingObjectLimit(object):
	def __init__(self, obj, key, val):
		self.obj = obj
		self.key = key
		assert len(val) == 1
		assert len(val[0]) == 6
		self.val = val[0]

	def _do_property(idx, doc, factor=1):
		def _getxfer(self):
			return self.val[idx]*factor
		def _setxfer(self, newval):
			l = list(self.val)
			l[idx] = newval / factor
			self.obj._change_limits(**{self.key: [tuple(l)]})
		return property(_getxfer, _setxfer, doc=doc)

	pps = _do_property(1, "Maximum Packets Per Second, 0 for unlimited")
	bps = _do_property(2, "Maximum Bits Per Second, 0 for unlimited", 8000)
	cps = _do_property(3, "Maximum New Connections Per Second, 0 for unlimited")
	queue_goal = _do_property(4, "Queue Goal in milliseconds, 0 for automatic")
	queue_size = _do_property(5, "Queue Size in packets, 0 for automatic")

	def __str__(self):
		r = []
		for n in ("pps", "bps", "cps", "queue_goal", "queue_size"):
			v = getattr(self, n)
			if v:
				r.append("%s=%d" % (n, v))
		if r:
			res = ", ".join(r)
		else:
			res = "Unlimited"

		return "<%s>" % res
	__repr__ = __str__

class ShapingObjectLimitGroup(object):
	def __init__(self, obj):
		self.obj = obj

	inbound = property(lambda self: ShapingObjectLimit(self.obj, "inbound", self.obj._prop.inbound))
	outbound = property(lambda self: ShapingObjectLimit(self.obj, "outbound", self.obj._prop.outbound))
	bidir = property(lambda self: ShapingObjectLimit(self.obj, "bidir", self.obj._prop.bidir))

	def __str__(self):
		return "<inbound=%s, outbound=%s, bidir=%s>" % (self.inbound, self.outbound, self.bidir)
	__repr__ = __str__



_Prop_ShapingObject = plutil.Prop('name', 'vbs_hours', 'inbound', 'outbound', 'bidir', 'split', 'max_connections', 'subscriber_netobject',
				  'id', 'creation_date', 'is_global', 'modification_date', 'creator', 'modified_by', 'systemid',
				  flags = ShapingObjectFlags)
class ShapingObject(object):
	"""
	Base class for ShapingObjects.

	This class should not be instantiated and should instead be created and
	accessed through:

		- L{Ruleset.shapingobject_add}
		- L{Ruleset.shapingobject_find}
		- L{Ruleset.shapingobject_find_id}
		- L{Ruleset.shapingobject_list}
		- L{Ruleset.shapingobject_remove}

	Class variables are read-only, use methods to alter the object.

	@type bidir: list
	@type outbound: list
	@type inbound: list
	@type name: str
	@type split: int
	@type vbs_hours: int
	@type max_connections: int
	@type creation_date: str
	@type creator: str
	@type id: int
	@type systemid: str
	@type modification_date: str
	@type modified_by: str
	@type is_global: C{boolean}
	@type flags: object
	@type machineid: str
	"""
	__slots__ = ('_prop', '_pldb')

	def __init__(self, _pldb=None, **props):
		self._pldb = _pldb
		self._prop = _Prop_ShapingObject(**props)

	def __repr__(self):
		return "<PacketLogic ShapingObject ('%s', limits=%s, split=%d, max_connections=%d, subscriber_netobject=%d, flags=%s)>" % (
			self._prop.name, self.limits, self._prop.split,
			self._prop.max_connections, self._prop.subscriber_netobject, str(list(self._prop.flags)))

	def change_limits(self, inbound=None, outbound=None, bidir=None):
		"""Deprecated function to change the limits in the shaping object. Any
		unspecified limits are left as is.

		@type inbound: C{list} of C{tuple}
		@param inbound: One entry list of containing the limit for incoming traffic, where
		                limit is (max_transfer, max_packets,
		                max_speed) max_transfer must be 0
		                (unlimited). Speed is specified in
		                kbyte/s.

		@type outbound: C{list} of C{tuple}
		@param outbound: same as inbound but for outgoing traffic

		@type bidir: C{list} of C{tuple}
		@param bidir: same as inbound but for both directions

		@attention: VBS is configured using attributes, therefore
		            the limit lists may only contain one entry
		            and specify max_transfer as 0.

		@attention: This method is deprecated, instead use the
		            L{limits} property to modify limits.

		"""

		import warnings
		warnings.warn("Use of change_limits is deprecated, use the limits property instead",
		              DeprecationWarning, stacklevel=1)
		self._change_limits(inbound, outbound, bidir)

	def _change_limits(self, inbound=None, outbound=None, bidir=None):
		if inbound is None:
			inbound = self._prop.inbound
		elif len(inbound) != 1:
			raise ValueError("Limit list must be of length 1 (inbound)")
		elif inbound[-1][0] != 0:
			raise ValueError("last limit must have max_transfer=0 (inbound)")
		else:
			inbound = _makenewformat(inbound)

		if outbound is None:
			outbound = self._prop.outbound
		elif len(outbound) != 1:
			raise ValueError("Limit list must be of length 1 (outbound)")
		elif outbound[-1][0] != 0:
			raise ValueError("last limit must have max_transfer=0 (outbound)")
		else:
			outbound = _makenewformat(outbound)

		if bidir is None:
			bidir = self._prop.bidir
		elif len(bidir) != 1:
			raise ValueError("Limit list must be of length 1 (bidir)")
		elif bidir[-1][0] != 0:
			raise ValueError("last limit must have max_transfer=0 (bidir)")
		else:
			bidir = _makenewformat(bidir)

		_shapingobjectupdate(self._pldb, self, self._prop.name, self._prop.vbs_hours, inbound, outbound, bidir,
				     self._prop.split, self._prop.max_connections, self._prop.subscriber_netobject, int(self._prop.flags))

		self._prop.inbound = inbound
		self._prop.outbound = outbound
		self._prop.bidir = bidir

	def change(self, split=None, max_connections=None, split_argument=None, subscriber_netobject=None): # inbound=None, outbound=None, bidir=None,
		"""
		Change properties of the shaping object. Any unspecified
		(or set to C{None}) property will be left as is.

		@type split: C{int}
		@param split: Type of split, one of: L{Ruleset <ruleset.Ruleset>}.SPLIT_NONE,
		              SPLIT_LOCALHOST, SPLIT_HOST_NETOBJECT,
		              SPLIT_LOCAL_NETOBJECT, SPLIT_SERVER_NETOBJECT,
		              SPLIT_CLIENT_NETOBJECT, SPLIT_CONNECTION,
			      SPLIT_SUBSCRIBER, SPLIT_LOCAL_NETWORKPREFIX

		@type max_connections: C{int}
		@param max_connections: Limit number of connections, or 0 for unlimited.

		@type split_argument: C{int}
		@param split_argument: NetObject with subscriber names
				       used to split this
				       ShapingObject with
				       SPLIT_SUBSCRIBER or prefix
				       length if
				       SPLIT_LOCAL_NETWORKPREFIX

		@type subscriber_netobject: C{int}
		@param subscriber_netobject: Deprecated alias for split_argument.

		@raises PLDBUnsupportedInFirmware: If split_argument argument is used and firmware doesn't support them.

		@attention: split_argument parameter requires v13.1 firmware or newer.
		"""

		#if inbound is None:
		inbound = self._prop.inbound
		#if outbound is None:
		outbound = self._prop.outbound
		#if bidir is None:
		bidir = self._prop.bidir

		vbs_hours = self._prop.vbs_hours

		if split is None:
			split = self.split
		elif split is NONEXISTANT_SPLIT_TYPE:
			raise PLDBUnsupportedInFirmware("Specified split is not supported in this firmware version")

		if max_connections is None:
			max_connections = self.max_connections

		if subscriber_netobject is not None:
			if split_argument is not None:
				raise ValueError, "Can't specify both split_argument and deprecated alias subscriber_netobject"
			split_argument = subscriber_netobject
		if split_argument is None:
			split_argument = self.split_argument

		_shapingobjectupdate(self._pldb, self, self._prop.name, vbs_hours, inbound, outbound, bidir,
				     split, max_connections, split_argument, int(self._prop.flags))

		#self._prop.inbound = inbound
		#self._prop.outbound = outbound
		#self._prop.bidir = bidir
		self._prop.vbs_hours = vbs_hours
		self._prop.split = split
		self._prop.max_connections = max_connections
		self._prop.subscriber_netobject = split_argument

	def set_flags(self, *reset, **flags):
		"""
		Set or reset one or more of the flags of the shaping object.
		You may also reset all flags with the single argument None, or clone an existing flags object.
		You may combine them.

		If only keyword argument(s) are used, only the specified flags will be affected.
		For a list of valid flags see the L{ShapingObjectFlags} class.
		Use a trueish value to set the flag, or falseish value to reset it.

			>>> shobj.set_flags(brown=True, host_fairness=False)
			>>> shobj.flags
			ShapingObjectFlags(brown=True)
			
			>>> shobj.set_flags(None)
			>>> shobj.flags
			ShapingObjectFlags(None)

			>>> shobj.set_flags(another_shobj.flags)
			>>> shobj.flags
			ShapingObjectFlags(host_fairness=True)
			
			>>> shobj.set_flags(None, brown=True)
			>>> shobj.flags
			ShapingObjectFlags(brown=True)
			
		"""
		if len(reset) == 1 and isinstance(reset[0], ShapingObject):
			# Might as well support supplying a shapingobject as argument too.
			self._prop.flags._set_flags(reset[0].flags, **flags)

		else:
			self._prop.flags._set_flags(*reset, **flags)

		_shapingobjectupdate(self._pldb, self, self._prop.name, self._prop.vbs_hours, self._prop.inbound, self._prop.outbound,
				     self._prop.bidir, self._prop.split, self._prop.max_connections, self._prop.subscriber_netobject, int(self._prop.flags))

	def rename(self, newname):
		"""
		Rename the current object

		I{destroys netobject cache}

		@type newname: C{str}
		@param newname: The name that the object should have after rename process

		"""

		_rename(self._pldb, self, newname)
		self._prop.name = newname

		self._pldb._invalidate_cache()

	def set_wfq(self, w1, w2, w3, w4, w5, w6):
		"""
		Set weights used for weighted fair queueing.

		@type w1: C{int}
		@param w1: Weight (in percent) to use for priority 4
		@type w2: C{int}
		@param w2: Weight (in percent) to use for priority 5
		@type w3: C{int}
		@param w3: Weight (in percent) to use for priority 6
		@type w4: C{int}
		@param w4: Weight (in percent) to use for priority 7
		@type w5: C{int}
		@param w5: Weight (in percent) to use for priority 8
		@type w6: C{int}
		@param w6: Weight (in percent) to use for priority 9

		@attention: wfq requires v14.0 firmware or newer.
		"""
		raise PLDBUnsupportedInFirmware("WFQ requires firmware v14.0 or newer")

	def _attr_set(self, key, value):
		if self._prop.is_global:
			raise ValueError("Can't set attributes on nonlocal objects")
		ci = self._pldb.session.shapingobject_attr_set(self._prop.id, key, value)
		c = plutil._AddCB(self._pldb)
		plutil._do(self._pldb, ci, c)

	def _attr_remove(self, key):
		if self._prop.is_global:
			raise ValueError("Can't remove attributes from nonlocal objects")
		ci = self._pldb.session.shapingobject_attr_remove(self._prop.id, key)
		c = plutil._AddCB(self._pldb)
		plutil._do(self._pldb, ci, c)

	def _attr_list(self):
		ci = self._pldb.session.shapingobject_attr_list(self._prop.systemid, self._prop.id)
		c = plutil._SimpleListCB(self._pldb, self._pldb.session.shapingobject_attr_list_handle)
		c.list = []
		res = dict(plutil._do(self._pldb, ci, c))
		return plutil.AttributesDict(res, self._attr_set, self._attr_remove)
		
	def _machineid(self):
		""" Use of property 'machineid' is deprecated since v11.0beta5, use 'systemid' instead."""
		import warnings
		warnings.warn("Use of property 'machineid' is deprecated, use 'systemid' instead.",
			      DeprecationWarning, stacklevel=2)
		return self._prop.systemid
	
	machineid = property(_machineid)

	name = property(lambda self: self._prop.name, doc="Name of the shaping object")
	_limits_doc = """
	Limits enforced by this shapingobject.

	    >>> o.limits
	    <inbound=<Unlimited>, outbound=<Unlimited>, bidir=<Unlimited>>
	    >>> o.limits.inbound.bps = 128000
	    >>> o.limits.inbound.queue_goal = 100
	    >>> o.limits
	    <inbound=<bps=128000, queue_goal=100>, outbound=<Unlimited>, bidir=<Unlimited>>

	@raises PLDBUnsupportedInFirmware: If accessed without firmware support.

 	@attention: This (new style) access for limits is only available in 12.4 and newer.
"""
	limits = property(lambda self: ShapingObjectLimitGroup(self), doc=_limits_doc)

	_oldstyle_limits_doc = """
	List of %s shaping limits.

	Each entry in the list is a tuple of (max_transfer,
	max_packets, max_speed). Same format used in
	L{shapingobject_add <ruleset.Ruleset.shapingobject_add>}.

	@attention: This format is deprecated, instead use the L{limits} property.
"""
	inbound = property(lambda self: _makeoldformat(self._prop.inbound, "inbound"), doc=_oldstyle_limits_doc % "inbound")
	outbound = property(lambda self: _makeoldformat(self._prop.outbound, "outbound"), doc=_oldstyle_limits_doc % "outbound")
	bidir = property(lambda self: _makeoldformat(self._prop.bidir, "bidir"), doc=_oldstyle_limits_doc % "bidirectional")
	split = property(lambda self: self._prop.split, doc="Split by type used by this object. One of the SPLIT_* constants available on the L{Ruleset object <ruleset.Ruleset>}")
	max_connections = property(lambda self: self._prop.max_connections, doc="Maximum number of allowed connections")
	flags = property(lambda self: self._prop.flags, doc="L{ShapingObjectFlags} object containing the flags (if any) set on this ShapingObject.")
	id = property(lambda self: self._prop.id, doc="id number of this object")
	creation_date = property(lambda self: self._prop.creation_date, doc="Date and time when this object was created")
	is_global = property(lambda self: self._prop.is_global, doc="True if this object resides on another system")
	modification_date = property(lambda self: self._prop.modification_date, doc="Date and time when this object was last modified")
	creator = property(lambda self: self._prop.creator, doc="Username that created this object")
	modified_by = property(lambda self: self._prop.modified_by, doc="Username that last modified this object")
	systemid = property(lambda self: self._prop.systemid, doc="System ID of the PacketLogic this object belongs to. (Replaces machineid)")
	split_argument = property(lambda self: self._prop.subscriber_netobject, doc="NetObject with subscriber names used to split this ShapingObject with SPLIT_SUBSCRIBER or prefix length if SPLIT_LOCAL_NETWORKPREFIX")
	subscriber_netobject = property(lambda self: self.split_argument, doc="NetObject with subscriber names used to split this ShapingObject with split-by-subscriber (0 when not relevant for the splittype)")
	wfq = property(lambda self: (100, 0, 0, 0, 0, 0), doc="Weights (in percent) to use for weighted fair queueing.")

	_attributes_doc = """
	Arbitrary key/value attributes stored for this shapingobject.

	    >>> o.attributes
	    {}
	    >>> o.attributes['test'] = "Arbitrary string value"
	    >>> o.attributes
	    {'test': "Arbitrary string value"}

	@raises PLDBUnsupportedInFirmware: If attributes is accessed without firmware support.

 	@attention: Attributes requires v12.2 firmware or newer.
	"""
	attributes = property(_attr_list, doc=_attributes_doc)

