"""
This module contains functionality for accessing the realtime parts of
the PacketLogic system, such as system diagnostics, channel statistics
and dynamic netobject items.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> rt = pl.Realtime()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from thisversion import deprecatedalias, plrcd, packetlogicd, SysDiag

from packetlogic2.exceptions import PLDUnsupportedInFirmware

import time
import select


##
## Here be stuff for generating proxy functions.  These functions
## allow us to give the PLD object the same functions that earlier
## versions (v13.0 and earlier) exported, with same arguments and
## docstring.
##

_PROXY_FUNC_TEMPLATE="""
def %(name)s (%(args)s):
	return self.%(subattr)s.%(name)s(%(callargs)s)
"""

def _generate_args(a, iscall):
	import inspect
	res = []
	args,varargs,kwargs,defaults = inspect.getargspec(a)
	assert(args[0] == 'self')
	if iscall:
		args = args[1:]
	if defaults:
		nondefaultargs = args[:len(args)-len(defaults)]
	else:
		nondefaultargs = args
	for a in nondefaultargs:
		if isinstance(a, list):
			res.append('(%s)' % (",".join(a)))
		else:
			res.append(a)
	if defaults:
		defaultargs = args[len(args)-len(defaults):]
		for a,v in zip(defaultargs, defaults):
			if iscall:
				res.append("%s" % a)
			else:
				res.append("%s=%s" % (a, repr(v)))
	if varargs:
		res.append("*" + varargs)
	if kwargs:
		res.append("**" + kwargs)

	return ",".join(res)

def _do_generic_func(name, cls, subattr):
	actualfunc = getattr(cls, name)
	proxyfunc = _PROXY_FUNC_TEMPLATE % {
		'name': name,
		'args': _generate_args(actualfunc, False),
		'callargs': _generate_args(actualfunc, True),
		'subattr': subattr}
	exec_globals = {}
	exec proxyfunc in exec_globals
	func = exec_globals[name]
	func.__doc__ = actualfunc.__doc__
	return func

def _do_plrcdfunc(name):
	return _do_generic_func(name, plrcd.PLRCD, "_plrcd")
def _do_packetlogicdfunc(name):
	return _do_generic_func(name, packetlogicd.PacketLogicd, "_packetlogicd")


class PLD(object):
	"""
	Class for connecting and communicating with packetlogicd,
	the realtime part of PacketLogic.

	When using the multiversion API you will access this class via the
	L{PLConnection.Realtime() <packetlogic2.PLConnection.Realtime>}
	method.

	All functions that communicate with the PacketLogic Realtime
	system might raise errors common to all or many of the
	functions.Here is a list of such errors, that might not be
	listed on each method:

	@raises ValueError: As the superclass of an internal BadMessageError,
	if a malformed message is recieved on the PacketLogic connection.
	That normally will never happen though.

	@raises Realtime.PLDPermissionError: If the user lacks permission
	        for the requested action.

	@raises IOError: If a generic I/O error occurs while reading or
	        writing on the PacketLogic Realtime connection.

	@group Dynamic NetObject items: dyn_add,dyn_list,dyn_list_full,dyn_list_mask,dyn_list_no,dyn_remove
	@group Firewall Log: fwlog_clear,fwlog_query_offsets,fwlog_query_rule,fwlog_query_time_range
	@group Volume Based Shaping: vbs_list,vbs_query,vbs_reset,vbs_set
	@group Internal: dump_bgptable,dump_drdlstats,inject_packet,reload_drdl
	@group Realtime Data: add_channels_callback,add_netobj_callback,add_service_callback,add_shapingcnt_callback,add_shapingobject_callback,add_sysdiag_callback,add_stats_callback,add_update_done_callback,get_channels_data,get_netobj_data,get_services_data,get_sysdiag_data,get_stats_data,shapingcnt_request_all,stop_updating,update_done_callbacks,update_forever
	
	"""
	
	SUPPORTS_TIMEOUT = True # Actually it is pldb.Ruleset that is checked for this flag, but it seems apropriate to set this too.
	
	def __init__(self, host, user, password, port=42001, timeout=60.0):
		"""
		Initialize and connect to packetlogicd.

		@type host: C{str}
		@param host: A hostname or IP address for the PacketLogic
		             system you want to connect to.

		@type user: C{str}
		@param user: The username you want to authenticate
		             with.  The system share users with the
		             client. You can create the user you want
		             to use with the client.

		@type password: C{str}
		@param password: The password to authenticate with.

		@type port: C{int}
		@param port: The default port is 42001 and normally
		             you don't have to change that.

		@type timeout: C{float} or C{None}
		@param timeout: The connection timeout in seconds. Default None.
			Set to None to use blocking mode during connect, with no specified timeout.
			(A system dependent timeout may or may not still occur after some time.)
			New in version v11.3beta3
		       
		@raises RuntimeError: If authorization fails
		@raises RuntimeError: If server and API version does not match
		@raises IOError: B{"Operation timed out"}: If operation times out without response.
		@raises IOError: If an error occures on the socket when attempting to connect.
		
		"""
		self.host = host
		self.port = port
		self.user = user
		self.timeout = timeout
		#self.timeout = timeout or 60.0 # FIXME Should we set this to the timeout argument too?
		self.password = password

		self.updating = False

		self._update_done_callbacks = []
		self._update_done_count = 0

		self._packetlogicd_val = None
		self._plrcd_val = None
		self._sysdiag_val = None

	def close (self):
		"""
		Immediately disconnect from packetlogic system.

		After this method has been called this object becomes
		useless.

		It may be used to force a disconnect when the garbage
		collector can't be trusted to do a timely disconnect.
		"""
		if self._packetlogicd_val is not None:
			self._packetlogicd_val.close()
			self._packetlogicd_val = None
		if self._plrcd_val is not None:
			self._plrcd_val.close()
			self._plrcd_val = None

	def _get_plrcd(self):
		if self._plrcd_val is None:
			self._plrcd_val = plrcd.PLRCD(self.host, self.user, self.password, timeout=self.timeout)
			self._plrcd_val.add_update_done_callback(self._update_done)
		return self._plrcd_val
	def _get_packetlogicd(self):
		if self._packetlogicd_val is None:
			self._packetlogicd_val = packetlogicd.PacketLogicd(self.host, self.user, self.password, timeout=self.timeout)
			self._packetlogicd_val.add_update_done_callback(self._update_done)
		return self._packetlogicd_val
	def _get_sysdiag(self):
		if self._sysdiag_val is None:
			self._sysdiag_val = SysDiag(self.host, self.user, self.password, timeout=self.timeout)
		return self._sysdiag_val

	def _update_done(self):
		self._update_done_count += 1

	_plrcd = property(_get_plrcd)
	_packetlogicd = property(_get_packetlogicd)
	_sysdiag = property(_get_sysdiag)

	dyn_add = _do_plrcdfunc('dyn_add')
	dyn_remove = _do_plrcdfunc('dyn_remove')
	dyn_set = _do_plrcdfunc('dyn_set')
	dyn_strerror = _do_plrcdfunc('dyn_strerror')
	dyn_list = _do_plrcdfunc('dyn_list')
	_dyn_doc = """
		Get a handle for dynamic netobject operations. See L{Dyn} object for details.
"""
	dyn = property(lambda self: self._plrcd.dyn, doc=_dyn_doc)
	dyn_list_full = _do_plrcdfunc('dyn_list_full')
	dyn_list_no = _do_plrcdfunc('dyn_list_no')
	dyn_list_no_recurse = _do_plrcdfunc('dyn_list_no_recurse')
	dyn_list_mask = _do_plrcdfunc('dyn_list_mask')

	vbs_reset = _do_plrcdfunc('vbs_reset')
	vbs_set = _do_plrcdfunc('vbs_set')
	vbs_list = _do_plrcdfunc('vbs_list')
	vbs_query = _do_plrcdfunc('vbs_query')

	def add_update_done_callback(self, func):
		"""
		Add a callback to call after each update is done.

		@type func: C{callable}
		@param func: function to call. The function is called
			     without any arguments

		"""
		self._update_done_callbacks.append(func)

	def update_forever(self, interval=5):
		"""
		Start updating.

		This method will enter a loop that requests updates
		every 5 seconds (or other interval specified by the
		interval argument). Updates are sent to the callback
		functions registered with methods named add_*_callback
		(among others add_service_callback,
		add_channels_callback and
		add_update_done_callback). To break out of the
		updating loop call the stop_updating method.

		@type interval: C{float}
		@param interval: Interval (seconds) at which updates are requested.

		@raises PLDPermissionError: If the user lacks permission for at least one of the registered callbacks

		"""

		parts = []
		if self._packetlogicd_val:
			parts.append(self._packetlogicd_val)
		if self._plrcd_val:
			parts.append(self._plrcd_val)

		if not parts:
			raise ValueError("No callbacks added?")

		self.updating = True

		next = time.time()
		while self.updating:
			inl = []
			outl = []
			for p in parts:
				p._fill_socketlist(inl,outl)

			now = time.time()
			i,o,e = select.select(inl, outl, inl, max(0, next-now))

			if len(i) == 0 and len(o) == 0 and len(e) == 0:
				# request new
				for p in parts:
					self._update_done_count = 0
					p._request_update()
				next = next + interval
				continue

			for p in parts:
				p._handle_socket_events(i,o,e)

			if self._update_done_count == len(parts):
				self._update_done_count = 0
				for func in self._update_done_callbacks:
					func()

	def stop_updating(self):
		"""
		Stop the updating and clear callback functions.
		"""
		self.updating = False
		if self._packetlogicd_val:
			self._packetlogicd_val.stop_updating()
		if self._plrcd_val:
			self._plrcd_val.stop_updating()

        def break_update_forever(self):
                """
                Stop the updating, but keep callback functions.

                Like L{stop_updating} but allows resuming updating by
                calling update_forever again.
                """
                self.updating = False

	add_stats_callback = deprecatedalias('add_stats_callback', 'add_sysdiag_callback')

	add_shapingcnt_callback = _do_plrcdfunc('add_shapingcnt_callback')
	shapingcnt_request_all = _do_plrcdfunc('shapingcnt_request_all')
	_request_debug_stats = _do_packetlogicdfunc('_request_debug_stats')
	_request_thread_stats = _do_packetlogicdfunc('_request_thread_stats')
	_request_thread_stats_all = _do_packetlogicdfunc('_request_thread_stats_all')
	add_channels_callback = _do_packetlogicdfunc('add_channels_callback')
	add_netobj_callback = _do_packetlogicdfunc('add_netobj_callback')
	add_host_callback = _do_packetlogicdfunc('add_host_callback')
	get_view_builder = _do_packetlogicdfunc('get_view_builder')
	add_conn_view_callback = _do_packetlogicdfunc('add_conn_view_callback')
	add_aggr_view_callback = _do_packetlogicdfunc('add_aggr_view_callback')

	add_service_callback = _do_packetlogicdfunc('add_service_callback')
	add_shapingobject_callback = _do_packetlogicdfunc('add_shapingobject_callback')
	get_channels_data = _do_packetlogicdfunc('get_channels_data')

	def get_sysdiag_data(self, maxage=5):
		"""
		Get system diagnostics data once.

		(Should not be used simultaneously to
		add_stats_callback and update_forever)

		@type maxage: C{int}
		@param maxage: Max age in seconds of cached information.
		               If longer time has passed since last call,
			       new data will be requested from PacketLogic.

		@raises PLDPermissionError: If the user lacks permission.

		@rtype: C{dict}
		@returns: Dict containing all system diagnostics data in
		          the form of nested dicts, the toplevel dict
		          contains the different system diagnostics
		          zones.

		@attention: For firmware version 13.1 and newer this
		            is a compatibility wrapper around
		            pl.SysDiag(). If support for earlier
		            firmwares is not needed using pl.SysDiag()
		            directly is preferred.
		"""
		zones = self._sysdiag.zone_list()

		for z in zones.itervalues():
			z.refresh()

		def convert(z):
			r = {}
			for v in z.valuedefs.itervalues():
				for val in v.values():
					if len(val.level) == 2 and val.systemid == self._sysdiag.systemid:
						r[v.name] = {'name': v.name,
							       'value': val.value,
							       'momvalue': val.rate,
							       'maxval': val.max,
							       'minval': val.min,
							       'maxlast': val.max_tstamp,
							       'minlast': val.max_tstamp,
							       'type': v.type,
							       'oid': "%s.%s" % (z.oid, v.id)}
						break
			return r

		return dict([(z.name, convert(z)) for z in zones.itervalues()])


	def add_sysdiag_callback(self, func):
		"""
		Add a callback to call with system diagnostics data
		while updating.

		@type func: C{callable}
		@param func: function to call. The function is called
		             with a huge dict containing all system
		             diagnostics data.

		@attention: This method only works in v13.0 and
		            earlier. In 13.1 and newer use the SysDiag
		            resource instead.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.
		"""
		raise PLDUnsupportedInFirmware("This method only works in v13.0 and earlier")

	add_stats_callback = deprecatedalias('add_stats_callback', 'add_sysdiag_callback')
	get_stats_data = deprecatedalias('get_stats_data', 'get_sysdiag_data')

	get_netobj_data = _do_packetlogicdfunc('get_netobj_data')
	get_services_data = _do_packetlogicdfunc('get_services_data')
	fwlog_clear = _do_packetlogicdfunc('fwlog_clear')
	fwlog_query_rule = _do_packetlogicdfunc('fwlog_query_rule')
	fwlog_query_offsets = _do_packetlogicdfunc('fwlog_query_offsets')
	fwlog_query_time_range = _do_packetlogicdfunc('fwlog_query_time_range')
	vcrc = _do_packetlogicdfunc('vcrc')
	dump_bgptable = _do_packetlogicdfunc('dump_bgptable')
	#dump_drdlstats = _do_plrcdfunc('dump_drdlstats')
	drdl_revision = _do_plrcdfunc('drdl_revision')
	reload_drdl = _do_plrcdfunc('reload_drdl')
	inject_packet = _do_packetlogicdfunc('inject_packet')
	force_statistics_write = _do_packetlogicdfunc('force_statistics_write')
	_set_debug = _do_packetlogicdfunc('_set_debug')
	wait_for_ruleset = _do_plrcdfunc('wait_for_ruleset')

	def current_ruleset(self):
		return self._plrcd.current_ruleset
	current_ruleset = property(current_ruleset)

	# PLRCD KeyVault
	_keyvault_doc = """
		Get a handle for KeyVault operations. See L{KeyVault} object for details.
"""
	keyvault = property(lambda self: self._plrcd.keyvault, doc=_keyvault_doc)

	def instanceid(self):
		return self._plrcd.instanceid
	instanceid = property(instanceid)
	def distversion(self):
		return self._plrcd.distversion
	distversion = property(distversion)
	def systemid(self):
		return self._plrcd.systemid
	systemid = property(systemid)

#StatsZone = packetlogicd.StatsZone
RealtimeConn = packetlogicd.RealtimeConn
RealtimeHost = packetlogicd.RealtimeHost
RealtimeNetObject = packetlogicd.RealtimeNetObject
RealtimeNetObjectTree = packetlogicd.RealtimeNetObjectTree
# for documentation consistency
Dyn = plrcd.Dyn
KeyVault = plrcd.KeyVault
ViewBuilder = packetlogicd.ViewBuilder
