"""
@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from thisversion import PLDB
from thisversion import PL
from thisversion import MSG_TYPE_VERSION, MSG_TYPE_AUTH, MSG_TYPE_ERROR, MSG_TYPE_CRITICAL, MSG_TYPE_OK, MSG_TYPE_DATA, MSG_TYPE_BCAST_COMMIT
from thisversion import negotiate_flicka_with_timeout

from shared import shared
import select
import socket
import plutil
import sys
import time

from protocols import PROTOCOLS
import protocolobject
from commitid import CommitID

from packetlogic2.exceptions import PLDBError, PLDBUnsupportedFeatureError, PLDBNewDataCommitted, PLDUnsupportedInFirmware

class Resource(object):
	"""
	This is a parent class for resource classes.

	To obtain a resource class instance when using the
	multiversion API you use the apropriate method on
	the L{PLConnection <packetlogic2.PLConnection>} object.

	If you use the single version API, you have similar
	methods in the L{pldb} package.

	Any method using a superclass of this resource to communicate with a
	PacketLogic system may raise errors that is common for all or many of
	the functions that uses the connection to the PacketLogic.
	Below is a list of those generic errors that can occur, and that are
	not documented specifically on each of the methods that can raise them.

	Also generic ValueError and TypeError raised due to invalid arguments
	and similar exceptions will usually not be documented.

	@raises IOError: B{"Couldn't write to server"}: If it fails to send a command.
	@raises IOError: B{"Disconnected"}: If connection was lost while reading data.
	@raises IOError: B{"%(syserror_nr)d"}: A system error occured while reading data.
	@raises IOError: B{"Unknown"}: For other undefined errors while reading data.

	@raises resource.PLDBError: B{"Error from PLDB server on current command: '%(error_msg)s'"}: For errors from the PLDB server on the PacketLogic.
	@raises resource.PLDBError: B{"Critical error from PLDB server on current command: '%(error_msg)s'"}: For critical errors from the PLDB server on the PacketLogic.
	@raises PLDBNewDataCommitted: B{"Exception containing the CommitID. Reached as PLDBNewDataCommitted.commitid"}: If another client committs data, making any uncommited changes on your connection, as well as the cache of read rules and objects, invalid.

	@raises NotImplementedError: Is normally never raised unless you use classes intended only for internal use, as super-classes.
	"""
	
	SUPPORTS_TIMEOUT = True # Checked with hasattr in multiver-api to know if kwarg timeout is supported in this version
	
	def __init__(self, host, user, password, port=42000, timeout=None):
		"""
		Initialize and connect to PLDB.

		@type host: C{str}
		@param host: A hostname or IP address for the PacketLogic system
			you want to connect to.

		@type user: C{str}
		@param user: The username you want to authenticate with. The system
			share users with the client. You can create the user you
			want to use with the client.
		
		@type password: C{str}
		@param password: The password to authenticate with.

		@type port: C{int}
		@param port: The default port is 42000 and normally you don't have to
			change that.
			
		@type timeout: C{float} or C{None}
		@param timeout: The connection timeout in seconds. Default None.
		       Set to None to use blocking mode during connect, with no specified timeout.
		       New in version v11.3beta3
		       
		@raises RuntimeError: B{"Bad version %(version)d"}: if version of this API does
		       not match the contacted PacketLogic system. Normally you will never see
		       this error if using the multiversion API.

		@raises RuntimeError: B{"The message was not OK or DATA"}: If an invalid response was
		       returned by the PacketLogic. Normally will never happen.
		"""
		self.host = host
		self.port = port
		self.user = user
		self.connection_timeout = timeout
		self.timeout = 60.0
		#self.timeout = timeout or 60.0 # FIXME Should we set this to the timeout argument too?
		self.password = password
		self.authed = False
		self._full_version = None
		self.version = None
		self.PLDBVersion = None
		self.systemid = None
		self.cbs = {}
		self.cids = {}
		self._handling_commit = False
		if ':' in host:
			self.s = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
		else:
			self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

		self.commitid_begin = None
		self.commitid_commit = None
		self.commitid_commit_bcast = None

		try:
			self.__setup()
		except:
			# always close the socket if the constructor fails
			# otherwise we might leak file descriptors.
			try:
				self.s.close()
			except:
				pass
			raise

		self._cached_services_list = None
		
	def __setup (self):
		self.s.setblocking(False)
		self.s.settimeout(self.connection_timeout)

		try:
			self.s.connect((self.host, self.port))
		except socket.error:
			typ, val, tb = sys.exc_info()
			raise IOError, val, tb

		self.s.setblocking(False) # In case timeout was set to None, this will have been reset to blocking above.
		io = PL.PLIO(self.s.fileno())

		negotiate_flicka_with_timeout(self.s, io, self.connection_timeout)

		self.handler = PLDB.PLDBAPIHandle(io)

		self._addtypecb(MSG_TYPE_VERSION, self._handle_version)
		self._addtypecb(MSG_TYPE_AUTH, self._handle_auth)
		self._addtypecb(MSG_TYPE_BCAST_COMMIT, self._handle_commit_broadcast)
		self.handler.auth(self.user, self.password)

		while not self.authed:
			self._process()

		c = self.handler.session_new()
		self.session = 0
		self._addcidcb(c, self._handle_session_new)

		while self.session == 0:
			self._process()

		self._do_begin(self._resource())

	"""
	INTERNAL
	"""

	def _has_feature(self, feature):
		""" Return True or False depending on if the feature is enabled.
		If fetures haven't been retrieved before, it will create a Ruleset
		connection and fetch them.
		"""
		if not hasattr(shared, 'features'):
			from ruleset import Ruleset
			rs = Ruleset(self.host, self.user, self.password, port=self.port, timeout=self.connection_timeout)
			shared.features = rs.features_get()

		return getattr(shared.features, feature)
		
	def _require_features(self, *features):
		""" If any of the listed featuires aren't enabled, an PLDBUnsupportedFeatureError is raised."""
		for feature in features:
			if not self._has_feature(feature):
				raise PLDBUnsupportedFeatureError
				
	def _invalidate_cache(self, typ=None):
		self._cached_services_list = None

	def _resource(self):
		raise NotImplementedError()

	def _process(self, breakcid = None):
		"""
		This is an internal function that will process events as they
		occur on the socket
		"""

		inp = [self.s]

		if self.handler.want_write():
			outp = [self.s]
		else:
			outp = []


		(i, o, e) = select.select(inp, outp, inp, self.timeout)

		if len(i) + len(o) + len(e) == 0:
			raise IOError("Operation timed out")

		if len(e) > 0:
			raise IOError("Select returned error")

		if len(o) > 0:
			try:
				self.handler.write()
			except socket.error:
				typ, val, tb = sys.exc_info()
				raise IOError, val, tb
			return

		if len(i) > 0:
			while True:
				try:
					m = self.handler.readmsg()
				except socket.error:
					typ, val, tb = sys.exc_info()
					raise IOError, val, tb
				
				if not m:
					self._handling_commit = False
					break

				typ = m.get_type()

				if typ != MSG_TYPE_BCAST_COMMIT:
					self._handling_commit = False
					
				elif self._handling_commit:
					# We are in the middle of rolling back a commit message
					# so don't do it again since we will recurse too much.
					continue
					
				cid = m.get_cid()

				if breakcid == cid:
					return m

				if typ == MSG_TYPE_ERROR:
					raise PLDBError("Error from PLDB server on current command: '%s'" % m.get_error())
				if typ == MSG_TYPE_CRITICAL:
					raise PLDBError("Critical error from PLDB server on current command: '%s'" % m.get_error())

				if self.cbs.has_key(typ):
					cb = self.cbs[typ]
					cb(m)
				else:
					cb = None
					if self.cids.has_key(cid):
						cb = self.cids[cid]
					if cb:
						if cb(m):
							""" 
							This means that the handler got the
							last message in this transaction
							remove it from the list
							"""
							self._delcidcb(cid)

	def _addtypecb(self, typ, cb):
		self.cbs[typ] = cb

	def _addcidcb(self, cid, cb):
		self.cids[cid] = cb

	def _delcidcb(self, cid):
		del self.cids[cid]


	"""
	HANDLERS
	"""

	def _handle_version(self, m):
		### @raises RuntimeError("Bad version %(version)d"): if version of this API does not match the contacted PacketLogic system
		self._full_version = self.handler.version_handle(m)
		self.version = self._full_version[:3]
		self.PLDBVersion = self.version[0]
		self.machineid = self.version[2]
		self.systemid = self.version[2]

	def _handle_auth(self, m):
		res = self.handler.auth_handle(m)
		if not res:
			raise RuntimeError("Couldn't log in")

		self.authed = True

	def _handle_commit_broadcast(self, m):
		self._handling_commit = True
		self.commitid_commit_bcast = CommitID(*self.handler.commit_bcast_handle(m))
		self.rollback()
		self._handling_commit = False
		raise PLDBNewDataCommitted("New data was committed, rolling back", self.commitid_commit_bcast)

	def _handle_begin(self, m):
		if m.get_type() == MSG_TYPE_OK:
			return True

		elif m.get_type() == MSG_TYPE_DATA:
			self.commitid_begin = CommitID(*self.handler.begin_handle(m))
			return False

		raise RuntimeError("The message was not OK or DATA")

	def _handle_commit(self, m):
		if m.get_type() == MSG_TYPE_OK:
			return True

		elif m.get_type() == MSG_TYPE_DATA:
			self.commitid_commit = CommitID(*self.handler.commit_handle(m))
			return False

		raise RuntimeError("The message was not OK or DATA")

	def _handle_session_new(self, msg):
		if msg.get_type() == MSG_TYPE_OK:
			return True

		elif msg.get_type() == MSG_TYPE_DATA:
			self.session = self.handler.session_new_handle(msg)
			return False

		raise RuntimeError("The message was not OK or DATA")



	def _do_begin(self, resource):
		ci = self.session.resource_begin(resource)
		self.commitid_begin = None

		self._addcidcb(ci, self._handle_begin)

		while self.commitid_begin is None:
			self._process()

		return self.commitid_begin;

	"""
	PUBLIC
	"""
	def close(self):
		"""Immediately disconnects the resource.

		After this method has been called this object becomes
		useless.

		It may be used to force a disconnect when the garbage
		collector can't be trusted to do a timely disconnect.
		"""
		self.session = None
		self.handler = None
		self.s.close()
		self.s = None
		try:
			self._invalidate_cache()
		except NotImplementedError:
			pass
		class ClosedResource(object):
			def __getattr__(self, n):
				raise IOError("Disconnected")
		self.__dict__.clear()
		self.__class__ = ClosedResource

	def _commitid(self):
		"""The L{CommitID} associated with this session. (14.0 or newer only)"""
		return self.commitid_begin
	commitid = property(_commitid)

	def xfb_flags_list(self):
		""" List all XFB (Transfer Behaviour) flags, that can be used in FlagObjects.
		
		@rtype: C{list} of C{str}
		@returns: List of XFB flags
		"""

		return ["Pseudo Unidirectional",
			"Unidirectional",
			"Bulky",
			"Download",
			"CBR streaming",
			"Streaming",
			"Interactive",
			"Initial",
			"Client is local",
			"Server is local",
			"Asymmetric",
			"Untracked",
			"Beginning",
			"Random looking",
			"Flowsynced",
			"Inbound",
			"Outbound",
			"Established",
			"Flowsync failed",
			"Active"
			]

	def protocols_dict(self):
		""" Returns a dictionary with name:nr and nr:name items for IP protocols.
		Can be used to e.g translate numerical protocol in connlog.search()
		result to protocol name.
		
		@rtype: C{dict} of C{str}:C{int} and C{int}:C{str}
		@returns: Dict with name:nr and nr:name items for IP protocols.

		"""
		# socket.getprotobyname will do one of these directions of translation (hopefully with the same
		# result), but python does not expose the getprotobynr function, for that we need this dict.
		# FIXME Should ProtocolObject use this instead of socket.getprotobyname to ensure consistency?
		
		return PROTOCOLS
		
	def protocol(self, protocol, strict=False):
		""" Translate protocol name OR number to tuple with name AND number.
		Can be used to e.g translate numerical protocol in connlog.search()
		result to protocol name.
		
		An unknown protocol will be translated to e.g '#7' or -1, unless strict
		is set to True, in which case ValueError is raised for unknown protocols.

		@type protocol: C{int}, C{str} or a L{ProtocolObject <protocolobject.ProtocolObject>} Item
		@param protocol: Name or number of an IP protocol
		
		@type strict: C{bool}
		@param strict: If C{True}, ValueError is raised for an unknown
		               protocol, if C{False} (default), an unknown protocol
			       translates to number -1 or name '#7' (where 7 is the
			       supplied unknown protocol number).
		
		@rtype: C{tuple} of (C{str}, C{int})
		@returns: Tuple with name and nr of specified protocol.

		"""
		if isinstance(protocol, protocolobject.ProtocolObject.Item):
			protocol = int(str(protocol))

		#Is this conversion good or bad?
		#elif isinstance(protocol, basestr) and protocol.isdigit():
		#	protocol = int(protocol)
			
		if strict and protocol not in PROTOCOLS:
			raise ValueError, "Unknown protocol: %s" % (protocol,)
			
		elif isinstance(protocol, int):
			return PROTOCOLS.get(protocol, '#%d' % protocol), protocol

		elif isinstance(protocol, basestring):
			return protocol, PROTOCOLS.get(protocol, -1)

		else:
			raise TypeError, "protocol should be either a string or an integer."
		
	def services_list(self):
		""" List all services
		
		@rtype: C{list} of C{str}
		@returns: List of services
		"""
		if self._cached_services_list is None:
			self._cached_services_list = plutil._do(self, self.session.services_list(), plutil._ListCB(self, self.session.services_list_handle, PLDB.PLServicesList, False))
		return self._cached_services_list[:]

	def services_prop_list(self):
		""" List all properties. 

		@rtype: C{list} of C{str}
		@returns: List of service properties
		"""
		return plutil._do(self, self.session.services_prop_list(), plutil._ListCB(self, self.session.services_list_handle, PLDB.PLServicesList, False))

	def services_cmd(self, services=[], props=[]):
		""" Send services to pldbd

		Sets stringtable for services and properties. Do B{NOT} use this unless you know exactly what you are doing.

		@type services: C{list} of C{str}
		@param services: List of services

		@type props: C{list} of C{str}
		@param props: List of service properties

		"""
		return plutil._do(self, self.session.services_cmd(services, props), plutil._AddCB(self))

	def commit(self, message="PythonAPI commit"):
		"""
		Store the current transaction to the database.

		When you do this PacketLogic will reload the
		ruleset. This will take a lot of CPU resources from
		the system, try to do this in batches.

		Observe that if you don't commit the changes you've
		made, no other resource (including the client) will
		see your changes.

		@type message: C{str}
		@param message: Message recorded in the commit log. Ignored in v12.1 and earlier.

		@attention: message parameter is ignored in v12.1 firmware and earlier.
		"""
		self._cached_services_list = None
		ci = self.session.resource_commit(message)
		self.commitid_commit = None

		self._addcidcb(ci, self._handle_commit)

		while self.commitid_commit is None:
			self._process()

		# Ignore self.commitid_commit in the current COMMIT reply.
		# We will get a new one in this following BEGIN
		return self._do_begin(self._resource())

	def rollback(self):
		"""
		Abort the current transaction and discard the data.

		This will also automatically happen if your connection
		drops or that you exit the program without commit.

		@raises resource.PLDBError: If the rollback command failed.
		"""
		ci = self.session.resource_rollback()
		m = None
		while not m:
			# we are rolling back, ignore any errors..
			try:
				m = self._process(ci)
			except PLDBError:
				pass

		self._invalidate_cache()
		if m.get_type() == MSG_TYPE_OK:
			self._do_begin(self._resource())
			return

		raise PLDBError("Error while rolling back")

	def ping(self, string=""):
		"""
		Send a command to server to test if it is up.
		"""
		plutil._do(self, self.session.ping(string), plutil._AddCB(self, self.session.ping_handle))

	def wait_for_commit(self, pinginterval=60):
		"""
		Wait for new data to be committed on resource.

		This method returns as soon as the server signals that
		another client has committed new data to the
		resource. It also periodically tests the connection to
		the server. If connection is broken (or any other
		error happens) an exception is raised.

		@type pinginterval: int
		@param pinginterval: Interval used to check if connection still is alive.
		"""
		nextping = time.time() + pinginterval
		ip = [self.s]
		while True:
			try:
				maxsleep = nextping - time.time()
				if maxsleep > 0:
					i,o,e = select.select(ip, [], ip, maxsleep)
					if i:
						self._process()
				self.ping()
				nextping = time.time() + pinginterval
			except PLDBNewDataCommitted:
				return
